@@ -4,7 +4,7 @@ from typing import Optional, Tuple
 import discord
 from discord.ext.commands import Bot
 
-from . import enums, log, node, player_manager
+from . import enums, log, node, player_manager, errors
 
 __all__ = [
     "initialize",
@@ -18,10 +18,11 @@ __all__ = [
     "register_stats_listener",
     "unregister_stats_listener",
     "all_players",
+    "all_connected_players",
     "active_players",
+    "wait_until_ready",
 ]
 
-
 _event_listeners = []
 _update_listeners = []
 _stats_listeners = []
@@ -32,11 +33,11 @@ async def initialize(
     bot: Bot,
     host,
     password,
-    rest_port,
     ws_port,
     timeout=30,
     resume_key: Optional[str] = None,
     resume_timeout: int = 60,
+    secured: bool = False,
 ):
     """
     Initializes the websocket connection to the lavalink player.
@@ -54,8 +55,6 @@ async def initialize(
         The hostname or IP address of the Lavalink node.
     password : str
         The password of the Lavalink node.
-    rest_port : int
-        The port of the REST API on the Lavalink node.
     ws_port : int
         The websocket port on the Lavalink Node.
     timeout : int
@@ -64,12 +63,13 @@ async def initialize(
         A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
     resume_timeout : inr
         How long the node should wait for a connection while disconnected before clearing all players.
+    secured: bool
+           Whether to use the `wss://` and `https://` protocol.
     """
     global _loop
     _loop = bot.loop
 
     player_manager.user_id = bot.user.id
-    player_manager.channel_finder_func = bot.get_channel
     register_event_listener(_handle_event)
     register_update_listener(_handle_update)
 
@@ -80,14 +80,16 @@ async def initialize(
         host,
         password,
         port=ws_port,
-        rest=rest_port,
         user_id=player_manager.user_id,
         num_shards=bot.shard_count if bot.shard_count is not None else 1,
         resume_key=resume_key,
         resume_timeout=resume_timeout,
+        bot=bot,
+        secured=secured,
     )
 
     await lavalink_node.connect(timeout=timeout)
+    lavalink_node._retries = 0
 
     bot.add_listener(node.on_socket_response)
     bot.add_listener(_on_guild_remove, name="on_guild_remove")
@@ -95,7 +97,7 @@ async def initialize(
     return lavalink_node
 
 
-async def connect(channel: discord.VoiceChannel):
+async def connect(channel: discord.VoiceChannel, deafen: bool = False):
     """
     Connects to a discord voice channel.
 
@@ -117,7 +119,7 @@ async def connect(channel: discord.VoiceChannel):
         If there are no available lavalink nodes ready to connect to discord.
     """
     node_ = node.get_node(channel.guild.id)
-    p = await node_.player_manager.create_player(channel)
+    p = await node_.player_manager.create_player(channel, deafen=deafen)
     return p
 
 
@@ -129,7 +131,7 @@ def get_player(guild_id: int) -> player_manager.Player:
 async def _on_guild_remove(guild):
     try:
         p = get_player(guild.id)
-    except (IndexError, KeyError):
+    except (errors.NodeNotFound, errors.PlayerNotFound):
         pass
     else:
         await p.disconnect()
@@ -147,7 +149,7 @@ def register_event_listener(coro):
     be a :py:class:`TrackEndReason`.
 
     If the second argument is :py:attr:`LavalinkEvents.TRACK_EXCEPTION`, the extra
-    will be an error string.
+    will be a dictionary with ``message``, ``cause``, and ``severity`` keys.
 
     If the second argument is :py:attr:`LavalinkEvents.TRACK_STUCK`, the extra will
     be the threshold milliseconds that the track has been stuck for.
@@ -184,7 +186,7 @@ def _get_event_args(data: enums.LavalinkEvents, raw_data: dict):
     try:
         node_ = node.get_node(guild_id, ignore_ready_status=True)
         player = node_.player_manager.get_player(guild_id)
-    except (IndexError, KeyError):
+    except (errors.NodeNotFound, errors.PlayerNotFound):
         if data != enums.LavalinkEvents.TRACK_END:
             log.debug(
                 "Got an event for a guild that we have no player for."
@@ -197,7 +199,12 @@ def _get_event_args(data: enums.LavalinkEvents, raw_data: dict):
     if data == enums.LavalinkEvents.TRACK_END:
         extra = enums.TrackEndReason(raw_data.get("reason"))
     elif data == enums.LavalinkEvents.TRACK_EXCEPTION:
-        extra = raw_data.get("error")
+        exception_data = raw_data["exception"]
+        extra = {
+            "message": exception_data["message"],
+            "cause": exception_data["cause"],
+            "severity": enums.ExceptionSeverity(exception_data["severity"]),
+        }
     elif data == enums.LavalinkEvents.TRACK_STUCK:
         extra = raw_data.get("thresholdMs")
     elif data == enums.LavalinkEvents.TRACK_START:
@@ -207,6 +214,7 @@ def _get_event_args(data: enums.LavalinkEvents, raw_data: dict):
             "code": raw_data.get("code"),
             "reason": raw_data.get("reason"),
             "byRemote": raw_data.get("byRemote"),
+            "channelID": player.channel.id if player.channel else None,
         }
     return player, data, extra
 
@@ -248,7 +256,7 @@ def register_update_listener(coro):
         _update_listeners.append(coro)
 
 
-async def _handle_update(player, data: enums.PlayerState, raw_data: dict):
+async def _handle_update(player, data: node.PlayerState, raw_data: dict):
     await player.handle_player_update(data)
 
 
@@ -257,7 +265,7 @@ def _get_update_args(data: enums.PlayerState, raw_data: dict):
 
     try:
         player = get_player(guild_id)
-    except KeyError:
+    except (errors.NodeNotFound, errors.PlayerNotFound):
         log.debug(
             "Got a player update for a guild that we have no player for."
             " This may be because of a forced voice channel disconnect."
@@ -339,13 +347,17 @@ def dispatch(op: enums.LavalinkIncomingOp, data, raw_data: dict):
         _loop.create_task(coro(*args))
 
 
-async def close():
+async def close(bot):
     """
     Closes the lavalink connection completely.
     """
+    log.debug("Closing Lavalink connections")
     unregister_event_listener(_handle_event)
     unregister_update_listener(_handle_update)
+    bot.remove_listener(node.on_socket_response)
+    bot.remove_listener(_on_guild_remove, name="on_guild_remove")
     await node.disconnect()
+    log.debug("All Lavalink nodes have been disconnected")
 
 
 # Helper methods
@@ -357,6 +369,33 @@ def all_players() -> Tuple[player_manager.Player]:
     return ret
 
 
+def all_connected_players() -> Tuple[player_manager.Player]:
+    nodes = node._nodes
+    ret = tuple(p for n in nodes for p in n.player_manager.players if p.connected)
+    return ret
+
+
 def active_players() -> Tuple[player_manager.Player]:
-    ps = all_players()
-    return tuple(p for p in ps if p.current is not None)
+    ps = all_connected_players()
+    return tuple(p for p in ps if p.is_playing)
+
+
+async def wait_until_ready(timeout: Optional[float] = None, wait_if_no_node: Optional[int] = None):
+    if wait_if_no_node:
+        for iteration in range(0, abs(wait_if_no_node), 1):
+            if not node._nodes:
+                await asyncio.sleep(1)
+            else:
+                break
+    if not node._nodes:
+        raise asyncio.TimeoutError
+    for result in await asyncio.gather(
+        *(
+            node_.wait_until_ready(timeout)
+            for node_ in node._nodes
+            if (not node_.ready) and not node_.state == node.NodeState.DISCONNECTING
+        ),
+        return_exceptions=True,
+    ):
+        if result is not None:
+            raise result
