@@ -1,24 +1,40 @@
+from __future__ import annotations
 import asyncio
-import contextlib
 import secrets
 import string
 from collections import namedtuple
-from typing import Awaitable, List, Optional, cast
+from typing import List, Optional
 
 import aiohttp
 import typing
 from discord.backoff import ExponentialBackoff
+from discord.ext.commands import Bot
 
-from . import log, socket_log
-from .enums import *
+from . import __version__, ws_discord_log, ws_ll_log
+from .enums import (
+    LavalinkIncomingOp,
+    NodeState,
+    LavalinkEvents,
+    LavalinkOutgoingOp,
+    DiscordVoiceSocketResponses,
+)
 from .player_manager import PlayerManager
 from .rest_api import Track
+from .errors import AbortingNodeConnection, NodeNotReady, NodeNotFound
 
-__all__ = ["Stats", "Node", "NodeStats", "get_node", "get_nodes_stats", "join_voice"]
+__all__ = [
+    "Stats",
+    "Node",
+    "NodeStats",
+    "get_node",
+    "get_nodes_stats",
+    "get_all_nodes",
+    "join_voice",
+]
 
-_nodes = []  # type: List[Node]
+_nodes: List[Node] = []
 
-PlayerState = namedtuple("PlayerState", "position time")
+PlayerState = namedtuple("PlayerState", "position time connected")
 MemoryInfo = namedtuple("MemoryInfo", "reservable used free allocated")
 CPUInfo = namedtuple("CPUInfo", "cores systemLoad lavalinkLoad")
 
@@ -56,8 +72,8 @@ class Stats:
 
 
 # Node stats related class below and how it is called is originally from:
-# https://github.com/PythonistaGuild/Wavelink/blob/master/wavelink/stats.py#L41
-# https://github.com/PythonistaGuild/Wavelink/blob/master/wavelink/websocket.py#L132
+# https://github.com/PythonistaGuild/Wavelink/blob/abba49e9806af3c50886f82054ea603129ad08b9/wavelink/stats.py#L41
+# https://github.com/PythonistaGuild/Wavelink/blob/abba49e9806af3c50886f82054ea603129ad08b9/wavelink/websocket.py#L132
 class NodeStats:
     def __init__(self, data: dict):
         self.uptime = data["uptime"]
@@ -81,6 +97,17 @@ class NodeStats:
         self.frames_nulled = frame_stats.get("nulled", -1)
         self.frames_deficit = frame_stats.get("deficit", -1)
 
+    def __repr__(self):
+        return (
+            "<NodeStats: "
+            f"uptime={self.uptime}, "
+            f"players={self.players}, "
+            f"playing_players={self.playing_players}, "
+            f"memory_free={self.memory_free}, memory_used={self.memory_used}, "
+            f"cpu_cores={self.cpu_cores}, system_load={self.system_load}, "
+            f"lavalink_load={self.lavalink_load}>"
+        )
+
 
 class Node:
 
@@ -94,11 +121,12 @@ class Node:
         host: str,
         password: str,
         port: int,
-        rest: int,
         user_id: int,
         num_shards: int,
         resume_key: Optional[str] = None,
         resume_timeout: int = 60,
+        bot: Bot = None,
+        secured: bool = False,
     ):
         """
         Represents a Lavalink node.
@@ -117,8 +145,6 @@ class Node:
             Password for the Lavalink player.
         port : int
             Port of the Lavalink player event websocket.
-        rest : int
-            Port for the Lavalink REST API.
         user_id : int
             User ID of the bot.
         num_shards : int
@@ -127,13 +153,15 @@ class Node:
             A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
         resume_timeout : int
             How long the node should wait for a connection while disconnected before clearing all players.
+        bot: AutoShardedBot
+            The Bot object that connects to discord.
         """
         self.loop = _loop
+        self.bot = bot
         self.event_handler = event_handler
         self.get_voice_ws = voice_ws_func
         self.host = host
         self.port = port
-        self.rest = rest
         self.password = password
         self._resume_key = resume_key
         if self._resume_key is None:
@@ -142,17 +170,21 @@ class Node:
         self._resuming_configured = False
         self.num_shards = num_shards
         self.user_id = user_id
+        self.secured = secured
 
-        self.ready = asyncio.Event()
+        self._ready_event = asyncio.Event()
 
         self._ws = None
         self._listener_task = None
         self.session = aiohttp.ClientSession()
+        self.reconnect_task = None
+        self.try_connect_task = None
 
         self._queue: List = []
 
         self.state = NodeState.CONNECTING
         self._state_handlers: List = []
+        self._retries = 0
 
         self.player_manager = PlayerManager(self)
 
@@ -167,6 +199,16 @@ class Node:
             aiohttp.WSMsgType.CLOSED,
         )
 
+    def __repr__(self):
+        return (
+            "<Node: "
+            f"state={self.state.name}, "
+            f"host={self.host}, "
+            f"port={self.port}, "
+            f"password={'*' * len(self.password)}, resume_key={self._resume_key}, "
+            f"shards={self.num_shards}, user={self.user_id}, stats={self.stats}>"
+        )
+
     @property
     def headers(self) -> dict:
         return self._get_connect_headers()
@@ -181,7 +223,7 @@ class Node:
             self._resume_key.__repr__()
             return self._resume_key
 
-    async def connect(self, timeout=None):
+    async def connect(self, timeout=None, *, shutdown=False):
         """
         Connects to the Lavalink player event websocket.
 
@@ -191,37 +233,30 @@ class Node:
             Time after which to timeout on attempting to connect to the Lavalink websocket,
             ``None`` is considered never, but the underlying code may stop trying past a
             certain point.
+        shutdown : bool
+            Whether the node was told to shut down
 
         Raises
         ------
         asyncio.TimeoutError
             If the websocket failed to connect after the given time.
+        AbortingNodeConnection
+            If the connection attempt must be aborted during a reconnect attempt
         """
-        self._is_shutdown = False
-
-        uri = "ws://{}:{}".format(self.host, self.port)
-
-        log.debug("Lavalink WS connecting to %s with headers %s", uri, self.headers)
-
-        for task in asyncio.as_completed([self._multi_try_connect(uri)], timeout=timeout):
-            with contextlib.suppress(Exception):
-                if await cast(Awaitable[Optional[aiohttp.ClientWebSocketResponse]], task):
-                    break
+        self._is_shutdown = shutdown
+        if self.secured:
+            uri = f"wss://{self.host}:{self.port}"
         else:
-            raise asyncio.TimeoutError
+            uri = f"ws://{self.host}:{self.port}"
 
-        log.debug("Creating Lavalink WS listener.")
-        if self._listener_task is not None:
-            self._listener_task.cancel()
-        self._listener_task = self.loop.create_task(self.listener())
-        if not self._resuming_configured:
-            self.loop.create_task(self._configure_resume())
-        if self._queue:
-            for data in self._queue:
-                await self.send(data)
-            self._queue.clear()
-        self.ready.set()
-        self.update_state(NodeState.READY)
+        ws_ll_log.info("Lavalink WS connecting to %s with headers %s", uri, self.headers)
+        if self.try_connect_task is not None:
+            self.try_connect_task.cancel()
+        self.try_connect_task = asyncio.create_task(self._multi_try_connect(uri))
+        try:
+            await asyncio.wait_for(self.try_connect_task, timeout=timeout)
+        except asyncio.CancelledError:
+            raise AbortingNodeConnection
 
     async def _configure_resume(self):
         if self._resuming_configured:
@@ -235,15 +270,19 @@ class Node:
                 )
             )
             self._resuming_configured = True
+            ws_ll_log.debug("Node Resuming has been configured.")
 
     async def wait_until_ready(self, timeout: Optional[float] = None):
-        await asyncio.wait_for(self.ready.wait(), timeout=timeout)
+        await asyncio.wait_for(self._ready_event.wait(), timeout=timeout)
 
     def _get_connect_headers(self) -> dict:
+        # Num-Shards is not used on Lavalink jar files >= v3.4
+        # but kept for compatibility to avoid NPEs on older builds
         headers = {
             "Authorization": self.password,
             "User-Id": str(self.user_id),
             "Num-Shards": str(self.num_shards),
+            "Client-Name": f"Red-Lavalink/{__version__}",
         }
         if self._resume_key:
             headers["Resume-Key"] = str(self._resume_key)
@@ -251,30 +290,65 @@ class Node:
 
     @property
     def lavalink_major_version(self):
-        if self.state != NodeState.READY:
-            raise RuntimeError("Node not ready!")
+        if not self.ready:
+            raise NodeNotReady("Node not ready!")
         return self._ws.response_headers.get("Lavalink-Major-Version")
 
+    @property
+    def ready(self) -> bool:
+        """
+        Whether the underlying node is ready for requests.
+        """
+        return self.state == NodeState.READY
+
     async def _multi_try_connect(self, uri):
         backoff = ExponentialBackoff()
         attempt = 1
+        if self._listener_task is not None:
+            self._listener_task.cancel()
+        if self._ws is not None:
+            await self._ws.close(code=4006, message=b"Reconnecting")
 
         while self._is_shutdown is False and (self._ws is None or self._ws.closed):
+            self._retries += 1
+            if self._is_shutdown is True:
+                ws_ll_log.error("Lavalink node was shutdown during a connect attempt.")
+                raise asyncio.CancelledError
             try:
                 ws = await self.session.ws_connect(url=uri, headers=self.headers, heartbeat=60)
-            except OSError:
+            except (OSError, aiohttp.ClientConnectionError):
+                if attempt > 5:
+                    raise asyncio.TimeoutError
                 delay = backoff.delay()
-                log.debug("Failed connect attempt %s, retrying in %s", attempt, delay)
+                ws_ll_log.warning("Failed connect attempt %s, retrying in %s", attempt, delay)
                 await asyncio.sleep(delay)
                 attempt += 1
             except aiohttp.WSServerHandshakeError:
-                return None
+                ws_ll_log.error("Failed connect WSServerHandshakeError")
+                raise asyncio.TimeoutError
             else:
+                if self._is_shutdown is True:
+                    ws_ll_log.error("Lavalink node was shutdown during a connect attempt.")
+                    raise asyncio.CancelledError
                 self.session_resumed = ws._response.headers.get("Session-Resumed", False)
                 if self._ws is not None and self.session_resumed:
-                    log.info(f"WEBSOCKET Resumed Session with key: {self._resume_key}")
+                    ws_ll_log.info("WEBSOCKET Resumed Session with key: %s", self._resume_key)
                 self._ws = ws
-                return self._ws
+                break
+        if self._is_shutdown is True:
+            raise asyncio.CancelledError
+        ws_ll_log.info("Lavalink WS connected to %s", uri)
+        ws_ll_log.debug("Creating Lavalink WS listener.")
+        if self._is_shutdown is False:
+            self._listener_task = self.loop.create_task(self.listener())
+            self.loop.create_task(self._configure_resume())
+            if self._queue:
+                temp = self._queue.copy()
+                self._queue.clear()
+                for data in temp:
+                    await self.send(data)
+            self._ready_event.set()
+            self.update_state(NodeState.READY)
 
     async def listener(self):
         """
@@ -284,47 +358,64 @@ class Node:
             msg = await self._ws.receive()
             if msg.type in self._closers:
                 if self._resuming_configured:
-                    self.loop.create_task(self._reconnect())
+                    if self.state != NodeState.RECONNECTING:
+                        if self.reconnect_task is not None:
+                            self.reconnect_task.cancel()
+                        ws_ll_log.info("[NODE] | NODE Resuming: %s", msg.extra)
+                        self.update_state(NodeState.RECONNECTING)
+                        self.reconnect_task = self.loop.create_task(
+                            self._reconnect(self._is_shutdown)
+                        )
                     return
                 else:
-                    log.debug("[NODE] | Listener closing: %s", msg.extra)
+                    ws_ll_log.info("[NODE] | Listener closing: %s", msg.extra)
                     break
             elif msg.type == aiohttp.WSMsgType.TEXT:
-                log.debug(f"[NODE] | Received Payload:: <{msg.data}>")
                 data = msg.json()
                 try:
                     op = LavalinkIncomingOp(data.get("op"))
                 except ValueError:
-                    socket_log.debug("[NODE] | Received unknown op: %s", data)
+                    ws_ll_log.verbose("[NODE] | Received unknown op: %s", data)
                 else:
-                    socket_log.debug("[NODE] | Received known op: %s", data)
+                    ws_ll_log.trace("[NODE] | Received known op: %s", data)
                     self.loop.create_task(self._handle_op(op, data))
             elif msg.type == aiohttp.WSMsgType.ERROR:
                 exc = self._ws.exception()
-                log.debug("[NODE] | Exception in WebSocket!", exc_info=exc)
+                ws_ll_log.warning(
+                    "[NODE] | An exception occurred on the websocket - Attempting to reconnect"
+                )
+                ws_ll_log.debug("[NODE] | Exception in WebSocket!", exc_info=exc)
                 break
             else:
-                log.debug(
+                ws_ll_log.debug(
                     "[NODE] | WebSocket connection received unexpected message: %s:%s",
                     msg.type,
                     msg.data,
                 )
-        log.debug(
-            "[NODE] | Listener exited: ws %s SHUTDOWN %s.", not self._ws.closed, self._is_shutdown
-        )
-        self.loop.create_task(self._reconnect())
+        if self.state != NodeState.RECONNECTING and not self._is_shutdown:
+            ws_ll_log.warning(
+                "[NODE] | %s - WS %s SHUTDOWN %s.", self, not self._ws.closed, self._is_shutdown
+            )
+            if self.reconnect_task is not None:
+                self.reconnect_task.cancel()
+            self.update_state(NodeState.RECONNECTING)
+            self.reconnect_task = self.loop.create_task(self._reconnect(self._is_shutdown))
 
     async def _handle_op(self, op: LavalinkIncomingOp, data):
         if op == LavalinkIncomingOp.EVENT:
             try:
                 event = LavalinkEvents(data.get("type"))
             except ValueError:
-                log.debug("Unknown event type: %s", data)
+                ws_ll_log.verbose("Unknown event type: %s", data)
             else:
                 self.event_handler(op, event, data)
         elif op == LavalinkIncomingOp.PLAYER_UPDATE:
             state = data.get("state", {})
-            state = PlayerState(position=state.get("position", 0), time=state.get("time", 0))
+            state = PlayerState(
+                position=state.get("position", 0),
+                time=state.get("time", 0),
+                connected=state.get("connected", False),
+            )
             self.event_handler(op, state, data)
         elif op == LavalinkIncomingOp.STATS:
             stats = Stats(
@@ -337,35 +428,66 @@ class Node:
             self.stats = NodeStats(data)
             self.event_handler(op, stats, data)
         else:
-            log.debug("Unknown op type: {}".format(data))
+            ws_ll_log.verbose("Unknown op type: %r", data)
 
-    async def _reconnect(self):
-        self.ready.clear()
+    async def _reconnect(self, shutdown: bool = False):
+        self._ready_event.clear()
 
-        if self._is_shutdown is True:
-            log.debug("Shutting down Lavalink WS.")
+        if self._is_shutdown is True or shutdown:
+            ws_ll_log.info("[NODE] | Shutting down Lavalink WS.")
             return
-
         if self.state != NodeState.CONNECTING:
             self.update_state(NodeState.RECONNECTING)
+        if self.state != NodeState.RECONNECTING:
+            return
+        backoff = ExponentialBackoff(base=1)
+        attempt = 1
+        while self.state == NodeState.RECONNECTING:
+            attempt += 1
+            if attempt > 10:
+                ws_ll_log.info("[NODE] | Failed reconnection attempt too many times, aborting ...")
+                asyncio.create_task(self.disconnect())
+                return
+            try:
+                await self.connect(shutdown=shutdown)
+            except AbortingNodeConnection:
+                return
+            except asyncio.TimeoutError:
+                delay = backoff.delay()
+                ws_ll_log.warning(
+                    "[NODE] | Lavalink WS reconnect attempt %s, retrying in %s",
+                    attempt,
+                    delay,
+                )
+                await asyncio.sleep(delay)
+            else:
+                ws_ll_log.info("[NODE] | Reconnect successful.")
+                self.dispatch_reconnect()
+                self._retries = 0
 
-        log.debug("Attempting Lavalink WS reconnect.")
-        try:
-            await self.connect()
-        except asyncio.TimeoutError:
-            log.debug("Failed to reconnect, please reinitialize lavalink when ready.")
-        else:
-            log.debug("Reconnect successful.")
+    def dispatch_reconnect(self):
+        for guild_id in self.player_manager.guild_ids:
+            self.event_handler(
+                LavalinkIncomingOp.EVENT,
+                LavalinkEvents.WEBSOCKET_CLOSED,
+                {
+                    "guildId": guild_id,
+                    "code": 42069,
+                    "reason": "Lavalink WS reconnected",
+                    "byRemote": True,
+                    "retries": self._retries,
+                },
+            )
 
     def update_state(self, next_state: NodeState):
         if next_state == self.state:
             return
 
-        log.debug(f"Changing node state: {self.state.name} -> {next_state.name}")
+        ws_ll_log.verbose("Changing node state: %s -> %s", self.state.name, next_state.name)
         old_state = self.state
         self.state = next_state
         if self.loop.is_closed():
-            log.debug("Event loop closed, not notifying state handlers.")
+            ws_ll_log.debug("Event loop closed, not notifying state handlers.")
             return
         for handler in self._state_handlers:
             self.loop.create_task(handler(next_state, old_state))
@@ -380,23 +502,37 @@ class Node:
     def unregister_state_handler(self, func):
         self._state_handlers.remove(func)
 
-    async def join_voice_channel(self, guild_id, channel_id):
+    async def join_voice_channel(self, guild_id, channel_id, deafen: bool = False):
         """
         Alternative way to join a voice channel if node is known.
         """
         voice_ws = self.get_voice_ws(guild_id)
-        await voice_ws.voice_state(guild_id, channel_id)
+        await voice_ws.voice_state(guild_id, channel_id, self_deaf=deafen)
 
     async def disconnect(self):
         """
         Shuts down and disconnects the websocket.
         """
+        global _nodes
         self._is_shutdown = True
-        self.ready.clear()
+        self._ready_event.clear()
+        self._queue.clear()
+        if (
+            self.try_connect_task is not None
+            and not self.try_connect_task.cancelled()
+            and not self.loop.is_closed()
+        ):
+            self.try_connect_task.cancel()
+        if (
+            self.reconnect_task is not None
+            and not self.reconnect_task.cancelled()
+            and not self.loop.is_closed()
+        ):
+            self.reconnect_task.cancel()
 
         self.update_state(NodeState.DISCONNECTING)
 
-        if self._resuming_configured:
+        if self._resuming_configured and not (self._ws is None or self._ws.closed):
             await self.send(dict(op="configureResuming", key=None))
         self._resuming_configured = False
         await self.player_manager.disconnect()
@@ -404,21 +540,27 @@ class Node:
         if self._ws is not None and not self._ws.closed:
             await self._ws.close()
 
-        if self._listener_task is not None and not self.loop.is_closed():
+        if (
+            self._listener_task is not None
+            and not self._listener_task.cancelled()
+            and not self.loop.is_closed()
+        ):
             self._listener_task.cancel()
 
         await self.session.close()
 
         self._state_handlers = []
-
-        _nodes.remove(self)
-        log.debug("Shutdown Lavalink WS.")
+        if len(_nodes) == 1:
+            _nodes = []
+        elif len(_nodes) > 1:
+            _nodes.remove(self)
+        ws_ll_log.info("Shutdown Lavalink WS.")
 
     async def send(self, data):
         if self._ws is None or self._ws.closed:
             self._queue.append(data)
         else:
-            log.debug("Sending data to Lavalink: %s", data)
+            ws_ll_log.trace("Sending data to Lavalink node: %s", data)
             await self._ws.send_json(data)
 
     async def send_lavalink_voice_update(self, guild_id, session_id, event):
@@ -434,22 +576,48 @@ class Node:
     async def destroy_guild(self, guild_id: int):
         await self.send({"op": LavalinkOutgoingOp.DESTROY.value, "guildId": str(guild_id)})
 
+    async def no_event_stop(self, guild_id: int):
+        await self.send({"op": LavalinkOutgoingOp.STOP.value, "guildId": str(guild_id)})
+
     # Player commands
     async def stop(self, guild_id: int):
-        await self.send({"op": LavalinkOutgoingOp.STOP.value, "guildId": str(guild_id)})
+        await self.no_event_stop(guild_id=guild_id)
         self.event_handler(
             LavalinkIncomingOp.EVENT, LavalinkEvents.QUEUE_END, {"guildId": str(guild_id)}
         )
 
-    async def play(self, guild_id: int, track: Track):
+    async def no_stop_play(
+        self,
+        guild_id: int,
+        track: Track,
+        replace: bool = True,
+        start: int = 0,
+        pause: bool = False,
+    ):
         await self.send(
             {
                 "op": LavalinkOutgoingOp.PLAY.value,
                 "guildId": str(guild_id),
                 "track": track.track_identifier,
+                "noReplace": not replace,
+                "startTime": str(start),
+                "pause": pause,
             }
         )
 
+    async def play(
+        self,
+        guild_id: int,
+        track: Track,
+        replace: bool = True,
+        start: int = 0,
+        pause: bool = False,
+    ):
+        # await self.send({"op": LavalinkOutgoingOp.STOP.value, "guildId": str(guild_id)})
+        await self.no_stop_play(
+            guild_id=guild_id, track=track, replace=replace, start=start, pause=pause
+        )
+
     async def pause(self, guild_id, paused):
         await self.send(
             {"op": LavalinkOutgoingOp.PAUSE.value, "guildId": str(guild_id), "pause": paused}
@@ -487,7 +655,7 @@ def get_node(guild_id: int, ignore_ready_status: bool = False) -> Node:
     for node in _nodes:
         guild_ids = node.player_manager.guild_ids
 
-        if ignore_ready_status is False and not node.ready.is_set():
+        if ignore_ready_status is False and not node.ready:
             continue
         elif len(guild_ids) < guild_count:
             guild_count = len(guild_ids)
@@ -497,7 +665,7 @@ def get_node(guild_id: int, ignore_ready_status: bool = False) -> Node:
             return node
 
     if least_used is None:
-        raise IndexError("No nodes found.")
+        raise NodeNotFound("No Lavalink nodes found.")
 
     return least_used
 
@@ -506,7 +674,11 @@ def get_nodes_stats():
     return [node.stats for node in _nodes]
 
 
-async def join_voice(guild_id: int, channel_id: int):
+def get_all_nodes() -> List[Node]:
+    return [node for node in _nodes]
+
+
+async def join_voice(guild_id: int, channel_id: int, deafen: bool = False):
     """
     Joins a voice channel by ID's.
 
@@ -516,8 +688,7 @@ async def join_voice(guild_id: int, channel_id: int):
     channel_id : int
     """
     node = get_node(guild_id)
-    voice_ws = node.get_voice_ws(guild_id)
-    await voice_ws.voice_state(guild_id, channel_id)
+    await node.join_voice_channel(guild_id, channel_id, deafen)
 
 
 async def disconnect():
@@ -532,13 +703,16 @@ async def on_socket_response(data):
     except ValueError:
         return
 
-    log.debug("Received Discord WS voice response: %s", data)
-
     guild_id = data["d"]["guild_id"]
 
     try:
         node = get_node(guild_id, ignore_ready_status=True)
-    except IndexError:
-        log.debug(f"Received unhandled socket response for guild: {guild_id}")
+    except NodeNotFound:
+        ws_discord_log.info(
+            f"Received unhandled Discord WS voice response for guild: %d, %s", int(guild_id), data
+        )
     else:
+        ws_ll_log.trace(
+            f"Received Discord WS voice response for guild: %d, %s", int(guild_id), data
+        )
         await node.player_manager.on_socket_response(data)
