@@ -176,6 +176,15 @@ class Track:
         """Overrides the default implementation"""
         return hash(tuple(sorted([self.track_identifier, self.title, self.author, self.uri])))
 
+    def __repr__(self):
+        return (
+            "<Track: "
+            f"track_identifier={self.track_identifier!r}, "
+            f"author={self.author!r}, "
+            f"length={self.length}, "
+            f"is_stream={self.is_stream}, uri={self.uri!r}, title={self.title!r}>"
+        )
+
 
 class LoadResult:
     """
@@ -268,8 +277,13 @@ class RESTClient:
 
     def __init__(self, node):
         self.node = node
+        self.secured = node.secured
         self._session = None
-        self._uri = "http://{}:{}/loadtracks?identifier=".format(node.host, node.rest)
+        if self.secured:
+            protocol = "https"
+        else:
+            protocol = "http"
+        self._uri = f"{protocol}://{node.host}:{node.port}/loadtracks?identifier="
         self._headers = {"Authorization": node.password}
 
         self.state = PlayerState.CONNECTING
@@ -298,7 +312,7 @@ class RESTClient:
                     },
                     "tracks": [],
                 }
-            log.debug(f"Received server disconnected error when player state = {self.state}")
+            log.debug("Received server disconnected error when player state = %s", self.state.name)
             raise
         return data
 
